/* filename: parse.y                 

Parser for phase 2
Modified by Nick Toothman
11/5/2009

*/
%{

#include <stdio.h>
#include <stdlib.h>
#include "symtab.h"

#define TRUE 1
#define FALSE 0
#define MAXTOKENLEN 40
#define MAXMEMSIZE  50 
#define MAX_MSG_LEN 50
#define YYDEBUG 1

typedef enum
{  Unknown, Number, Keyword, Identifier, Operator, Punctuator, StringLiteral }
TokenType;

%}

%token IF THEN ELSE END REPEAT UNTIL READ WRITE WHILE BREAK RETURN CLASS
%token NEW CONST VOID STRING
%token LTE LT GT GTE INEQ ASSIGN EQ PLUS MINUS TIMES OVER MOD COMMA
%token LPAREN LBRACE AND OR INC DECR CHARLIT RPAREN RBRACE SEMI NUM
%token ID INCR DECR
%token ENDFILE ERROR 

%% /* Simple BNF Grammar */

program     : CLASS ID block
            ;
block 	    : LBRACE statements RBRACE
	    ;
statements  : function
	    | statement
	    | function statements
	    | statement statements
	    ;
function    : fcall
	    | fdef
	    ;
fcall       : ID LPAREN list RPAREN
	    | VOID ID LPAREN list RPAREN 
            ;
fdef	    : ID LPAREN list RPAREN block
	    ;
list        : param
	    | param COMMA list
	    |
	    ;
param       : ID
	    | NUM
	    | STRING
	    | CHARLIT
	    ;
segment	    : statement
	    | fcall SEMI
	    | block
	    ;
statement   : ID ASSIGN expr SEMI 
	    | READ ID SEMI
	    | WRITE LPAREN expr RPAREN SEMI
	    | WHILE LPAREN expr RPAREN block
	    | IF LPAREN expr RPAREN segment
	    | IF LPAREN expr RPAREN segment ELSE segment
	    | NEW ID SEMI
	    | NEW ID ASSIGN expr SEMI
	    | BREAK SEMI
	    | CONST NEW ID ASSIGN expr SEMI
	    | fcall SEMI
	    | RETURN SEMI
	    | RETURN expr SEMI
	    ;
expr        : expr rel_op logical
	    | logical
	    ;
logical     : logical bool_op boolean
	    | boolean
	    ;
boolean     : boolean term_op term
	    | term
	    ;
term        : term fac_op factor
	    | factor
	    ;
factor      : LPAREN expr RPAREN
	    | ID
	    | NUM
	    | CHARLIT
	    | STRING
	    | un_op factor
	    | factor un_op
	    ;
rel_op      : AND | OR
	    ;
bool_op     : LTE 
	    | LT
	    | GT
	    | GTE
	    | INEQ
	    | EQ
	    ;
term_op     : PLUS
	    | MINUS
	    ;
fac_op      : TIMES
	    | OVER
	    | MOD
	    ;
un_op       : INCR
	    | DECR
	    ;
%%


void printToken( TokenType token, const char* tokenString );

extern FILE *yyin;
extern FILE *yyout;
extern int yyparse();

int yydebug = 1;
extern int line_no;
extern char tokenString[MAXTOKENLEN];

/* a quick and dirty pseudo-memory  */
char * memory[MAXMEMSIZE]; // an array of words, each 256 bytes */ 
int mem_offset = 0;  // stored in symbol table as an offset into pseudo-memory
void dump_mem ();     

int main( int argc,char *argv[] )
{
  int i;
  if(argc < 2) {
     printf("Usage: cf <source filename>\n");
     exit(0);
  }
  FILE *fp = fopen(argv[1],"r");
  if(!fp) {
    printf("Unable to open file for reading\n");
    exit(0);
  }
  yyin = fp;

  fp = fopen("sym.output","w");
  if(!fp) {
    printf("Unable to open file for writing\n");
    exit(0);
  }

  int flag = yyparse();

  dump_symtab(fp);   /* dump symbol table */
  dump_mem();

  /* cleanup */
  for (i = 0; i <= mem_offset; i++)
      free (memory[i]);
  fclose(fp);
  fclose(yyin);

  return flag;
}


yywrap()
{
  return(1);
}

int yyerror(char * message)
{ fprintf(stderr,"Syntax error at line %d: %s\n",line_no,message);
  fprintf(stderr,"Current token: ");
  printToken(yychar,tokenString);
  return 0;
}


/* not used but could be useful to display the parse tree */ 
void printToken( TokenType token, const char* tokenString )
{ switch (token)
  { case IF:
    case THEN:
    case ELSE:
    case END:
    case REPEAT:
    case UNTIL:
    case READ:
    case WRITE:
    case WHILE:
    case BREAK:
    case RETURN:
    case CLASS:
    case NEW:
    case CONST:
    case VOID:
      fprintf(stdout, "reserved word: %s\n",tokenString);
      break;
    case RPAREN: fprintf(stdout,")\n"); break;
    case LPAREN: fprintf(stdout,"(\n"); break;
    case LBRACE: fprintf(stdout,"{\n"); break;
    case RBRACE: fprintf(stdout,"}\n"); break;
    case TIMES: fprintf(stdout,"*\n"); break;
    case PLUS: fprintf(stdout,"+\n"); break;
    case NUM:
      fprintf(stdout, "NUM, val= %s\n",tokenString);
      break;
    case ID:
      fprintf(stdout, "ID, name= %s\n",tokenString);
      break;
    case ERROR:
      fprintf(stdout, "ERROR: %s\n",tokenString);
      break;
    default: 
      fprintf(stdout,"Unknown token: %d\n",tokenString);
  }
}

void dump_mem () {
   int i;
   for (i = 0; i <= mem_offset; i++)
      fprintf(stdout, "%s\n", memory[i]);
}

